#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/time.h>
#include <ctype.h>
#include "IOTCDevice.h"
#include "AVServer.h"
#include "VSaaS.h"
#include "P2PCam/AVFRAMEINFO.h"
#include "P2PCam/AVIOCTRLDEFs.h"
#include "NebulaError.h"
#include "playback.h"
#include "sample_constants.h"


#define ENABLE_RESEND 1
#define VSAAS_EVENT_TYPE 1 //0=live 1=record file
#define ENABLE_DTLS 0
#define SERVTYPE_STREAM_SERVER 0
#define FPS 30 // Video FPS
#define VIDEO_BUF_SIZE (1024 * 300)
#define VSAAS_CHANNEL_NUMBER 1 // For stage testing
#define VSAAS_CLOUD_SUPPORT_FLAG 1 //0 Not support；1 Support

char gAvAccount[NEW_MAXSIZE_VIEWACC + 1] = "admin";
char gAvPassword[NEW_MAXSIZE_VIEWPWD + 1] = "888888";
char gIotcAuthKey[IOTC_AUTH_KEY_LENGTH + 1] = "00000000"; //use "00000000" as default authkey
char gAvToken[MAX_TOKEN_LENGTH + 1] = "admin";			  //"AAAAAAAA";

char gVsaasInfoFilePath[] = "./vsaasinfo";
char gLiveIframePath[] = "./I_000.bin";
char gRecordFileName[] = "I_000.bin";
int gEnable_Vsaas = 0;
AV_Client gClientInfo[MAX_CLIENT_NUMBER];

static int gProcessRun = 1;
static int gOnlineNum = 0;
static char gUID[21];

static int gIOTCLoginSuccess = -1;

//########################################################
//#     Print error message
//########################################################
void PrintErrHandling(int nErr)
{
	switch (nErr)
	{
	case IOTC_ER_MASTER_NOT_RESPONSE:
		//-60 IOTC_ER_MASTER_NOT_RESPONSE
		printf("[Error code : %d]\n", IOTC_ER_MASTER_NOT_RESPONSE);
		printf("Master server doesn't respond.\n");
		printf("Please check the network wheather it could connect to the Internet.\n");
		break;
	case IOTC_ER_SERVER_NOT_RESPONSE:
		//-1 IOTC_ER_SERVER_NOT_RESPONSE
		printf("[Error code : %d]\n", IOTC_ER_SERVER_NOT_RESPONSE);
		printf("P2P Server doesn't respond.\n");
		printf("Please check the network wheather it could connect to the Internet.\n");
		break;
	case IOTC_ER_FAIL_RESOLVE_HOSTNAME:
		//-2 IOTC_ER_FAIL_RESOLVE_HOSTNAME
		printf("[Error code : %d]\n", IOTC_ER_FAIL_RESOLVE_HOSTNAME);
		printf("Can't resolve hostname.\n");
		break;
	case IOTC_ER_ALREADY_INITIALIZED:
		//-3 IOTC_ER_ALREADY_INITIALIZED
		printf("[Error code : %d]\n", IOTC_ER_ALREADY_INITIALIZED);
		printf("Already initialized.\n");
		break;
	case IOTC_ER_FAIL_CREATE_MUTEX:
		//-4 IOTC_ER_FAIL_CREATE_MUTEX
		printf("[Error code : %d]\n", IOTC_ER_FAIL_CREATE_MUTEX);
		printf("Can't create mutex.\n");
		break;
	case IOTC_ER_FAIL_CREATE_THREAD:
		//-5 IOTC_ER_FAIL_CREATE_THREAD
		printf("[Error code : %d]\n", IOTC_ER_FAIL_CREATE_THREAD);
		printf("Can't create thread.\n");
		break;
	case IOTC_ER_UNLICENSE:
		//-10 IOTC_ER_UNLICENSE
		printf("[Error code : %d]\n", IOTC_ER_UNLICENSE);
		printf("This UID is unlicense.\n");
		printf("Check your UID.\n");
		break;
	case IOTC_ER_NOT_INITIALIZED:
		//-12 IOTC_ER_NOT_INITIALIZED
		printf("[Error code : %d]\n", IOTC_ER_NOT_INITIALIZED);
		printf("Please initialize the IOTCAPI first.\n");
		break;
	case IOTC_ER_TIMEOUT:
		//-13 IOTC_ER_TIMEOUT
		break;
	case IOTC_ER_INVALID_SID:
		//-14 IOTC_ER_INVALID_SID
		printf("[Error code : %d]\n", IOTC_ER_INVALID_SID);
		printf("This SID is invalid.\n");
		printf("Please check it again.\n");
		break;
	case IOTC_ER_EXCEED_MAX_SESSION:
		//-18 IOTC_ER_EXCEED_MAX_SESSION
		printf("[Error code : %d]\n", IOTC_ER_EXCEED_MAX_SESSION);
		printf("[Warning]\n");
		printf("The amount of session reach to the maximum.\n");
		printf("It cannot be connected unless the session is released.\n");
		break;
	case IOTC_ER_CAN_NOT_FIND_DEVICE:
		//-19 IOTC_ER_CAN_NOT_FIND_DEVICE
		printf("[Error code : %d]\n", IOTC_ER_CAN_NOT_FIND_DEVICE);
		printf("Device didn't register on server, so we can't find device.\n");
		printf("Please check the device again.\n");
		printf("Retry...\n");
		break;
	case IOTC_ER_SESSION_CLOSE_BY_REMOTE:
		//-22 IOTC_ER_SESSION_CLOSE_BY_REMOTE
		printf("[Error code : %d]\n", IOTC_ER_SESSION_CLOSE_BY_REMOTE);
		printf("Session is closed by remote so we can't access.\n");
		printf("Please close it or establish session again.\n");
		break;
	case IOTC_ER_REMOTE_TIMEOUT_DISCONNECT:
		//-23 IOTC_ER_REMOTE_TIMEOUT_DISCONNECT
		printf("[Error code : %d]\n", IOTC_ER_REMOTE_TIMEOUT_DISCONNECT);
		printf("We can't receive an acknowledgement character within a TIMEOUT.\n");
		printf("It might that the session is disconnected by remote.\n");
		printf("Please check the network wheather it is busy or not.\n");
		printf("And check the device and user equipment work well.\n");
		break;
	case IOTC_ER_DEVICE_NOT_LISTENING:
		//-24 IOTC_ER_DEVICE_NOT_LISTENING
		printf("[Error code : %d]\n", IOTC_ER_DEVICE_NOT_LISTENING);
		printf("Device doesn't listen or the sessions of device reach to maximum.\n");
		printf("Please release the session and check the device wheather it listen or not.\n");
		break;
	case IOTC_ER_CH_NOT_ON:
		//-26 IOTC_ER_CH_NOT_ON
		printf("[Error code : %d]\n", IOTC_ER_CH_NOT_ON);
		printf("Channel isn't on.\n");
		printf("Please open it by IOTC_Session_Channel_ON() or IOTC_Session_Get_Free_Channel()\n");
		printf("Retry...\n");
		break;
	case IOTC_ER_SESSION_NO_FREE_CHANNEL:
		//-31 IOTC_ER_SESSION_NO_FREE_CHANNEL
		printf("[Error code : %d]\n", IOTC_ER_SESSION_NO_FREE_CHANNEL);
		printf("All channels are occupied.\n");
		printf("Please release some channel.\n");
		break;
	case IOTC_ER_TCP_TRAVEL_FAILED:
		//-32 IOTC_ER_TCP_TRAVEL_FAILED
		printf("[Error code : %d]\n", IOTC_ER_TCP_TRAVEL_FAILED);
		printf("Device can't connect to Master.\n");
		printf("Don't let device use proxy.\n");
		printf("Close firewall of device.\n");
		printf("Or open device's TCP port 80, 443, 8080, 8000, 21047.\n");
		break;
	case IOTC_ER_TCP_CONNECT_TO_SERVER_FAILED:
		//-33 IOTC_ER_TCP_CONNECT_TO_SERVER_FAILED
		printf("[Error code : %d]\n", IOTC_ER_TCP_CONNECT_TO_SERVER_FAILED);
		printf("Device can't connect to server by TCP.\n");
		printf("Don't let server use proxy.\n");
		printf("Close firewall of server.\n");
		printf("Or open server's TCP port 80, 443, 8080, 8000, 21047.\n");
		printf("Retry...\n");
		break;
	case IOTC_ER_NO_PERMISSION:
		//-40 IOTC_ER_NO_PERMISSION
		printf("[Error code : %d]\n", IOTC_ER_NO_PERMISSION);
		printf("This UID's license doesn't support TCP.\n");
		break;
	case IOTC_ER_NETWORK_UNREACHABLE:
		//-41 IOTC_ER_NETWORK_UNREACHABLE
		printf("[Error code : %d]\n", IOTC_ER_NETWORK_UNREACHABLE);
		printf("Network is unreachable.\n");
		printf("Please check your network.\n");
		printf("Retry...\n");
		break;
	case IOTC_ER_FAIL_SETUP_RELAY:
		//-42 IOTC_ER_FAIL_SETUP_RELAY
		printf("[Error code : %d]\n", IOTC_ER_FAIL_SETUP_RELAY);
		printf("Client can't connect to a device via Lan, P2P, and Relay mode\n");
		break;
	case IOTC_ER_NOT_SUPPORT_RELAY:
		//-43 IOTC_ER_NOT_SUPPORT_RELAY
		printf("[Error code : %d]\n", IOTC_ER_NOT_SUPPORT_RELAY);
		printf("Server doesn't support UDP relay mode.\n");
		printf("So client can't use UDP relay to connect to a device.\n");
		break;

	default:
		printf("[Unknow error code : %d]\n", nErr);
		break;
	}
}

//########################################################
//#     Fake VSaaS event for test
//########################################################
unsigned long gLastVsaasEventTimestamp_sec = 0;
void FakeEventForVsaas(unsigned long timestamp_sec)
{
    char att_json_str[2048] = {0}; /* A JSON data string use for create a Nebula object */
    unsigned long interval_ms = timestamp_sec - gLastVsaasEventTimestamp_sec;
    int ret = 0;
    NebulaJsonObject *attr_obj = NULL;  /* A nebula object as a parameter for call to avServNotifyCloudRecordStream() */

	/* Check IOTC Login status */
    if (gIOTCLoginSuccess == -1) {
        return;
	}

    if ((gLastVsaasEventTimestamp_sec == 0 || interval_ms > 60) && gEnable_Vsaas) // trigger fake event every 60sec
    {
        printf("\t%s() at %lu\n", __FUNCTION__, timestamp_sec);

        if (VSAAS_EVENT_TYPE) {
            // meaning the video is playback
            sprintf(att_json_str, "{\"starttime\":\"%lu\",\"protocol\":\"tutk\",\"event_id\":\"%d\",\"event_file\":\"%s\",\"media_type\":\"0\",\"channel\":%d}", timestamp_sec, VSAAS_EVENT_GENERAL, gRecordFileName, VSAAS_CHANNEL_NUMBER);
        } else {
            // meaning the video is live stream.
            sprintf(att_json_str, "{\"starttime\":\"live\",\"protocol\":\"tutk\",\"event_id\":\"%d\",\"media_type\":\"0\"}", VSAAS_EVENT_GENERAL);
        }

		/* Create a nebula object by a JSON data string */
    	ret = Nebula_Json_Obj_Create_From_String(att_json_str, &attr_obj);
		if(ret != NEBULA_ER_NoERROR) 
		{
			PrintErrHandling(ret);
            return;
		}
		
		printf("send avServNotifyCloudRecordStream\n");
	    /* Send a notification to Cloud for record streaming */
        ret = avServNotifyCloudRecordStream(attr_obj, 3000, NULL);
        printf("avServNotifyCloudRecordStream() =%d\n", ret);
        gLastVsaasEventTimestamp_sec = timestamp_sec;
		Nebula_Json_Obj_Release(attr_obj);

		if (ret != AV_ER_NoERROR) 
		{
            PrintErrHandling(ret);
            return;
        }
    }
}

//########################################################
//#     Callback function for avEnableVSaaS()
//########################################################
void VsaasConfigChangedCb(const char *VSaaSconfig)
{
	//save VSaaSconfig for avEnableVSaaS()
	FILE *vsaas_info_file_ptr = fopen(gVsaasInfoFilePath, "w+");
	if (vsaas_info_file_ptr == NULL)
	{
		printf("%s() Get VSaaS info:\n%s\n",__FUNCTION__, VSaaSconfig);
		printf("%s() write error!!\n", __FUNCTION__);
	}
	else
	{
		fwrite(VSaaSconfig, 1, strlen(VSaaSconfig), vsaas_info_file_ptr);
		fclose(vsaas_info_file_ptr);
	}

	while (gIOTCLoginSuccess < 0)
	{
		usleep(50000);
	}

}

void VSaaSUpdateContractInfoCb(const VSaaSContractInfo *contract_info)
{
    printf("%s() contract_type=%u\n", __FUNCTION__, contract_info->contract_type);
    printf("%s() event_recording_max_sec=%d\n", __FUNCTION__, contract_info->event_recording_max_sec);
    printf("%s() video_max_fps=%d\n", __FUNCTION__, contract_info->video_max_fps);
    printf("%s() recording_max_kbps=%d\n", __FUNCTION__, contract_info->recording_max_kbps);
    printf("%s() video_max_high=%d\n", __FUNCTION__, contract_info->video_max_high);
    printf("%s() video_max_width=%d\n", __FUNCTION__, contract_info->video_max_width);
	
	// todo: please use the contract info to adopt the stream bit rate or resolution
}

//########################################################
//#     Callback functions for avServStartEx()
//########################################################
int ExPasswordAuthCallBackFn(const char *account, char *pwd, unsigned int pwd_buf_size)
{
	if (strcmp(account, gAvAccount) != 0)
		return -1;

	if (pwd_buf_size <= strlen(gAvPassword))
		return -1;
	strcpy(pwd, gAvPassword);

	return 0;
}

int ExTokenAuthCallBackFn(const char *identity, char *token, unsigned int token_buf_size)
{
	if (token_buf_size <= strlen(gAvToken))
		return -1;
	strcpy(token, gAvToken);
	return 0;
}

int ExTokenDeleteCallBackFn(int av_index, const char *identity)
{
	//not implement in this sample
	return 0;
}

int ExTokenRequestCallBackFn(int av_index, const char *identity, const char *identity_description, char *token, unsigned int token_length)
{
	//not implement in this sample
	return 0;
}

void ExGetIdentityArrayCallBackFn(int avIndex, avServSendIdentityArray send_identity_array)
{
	//not implement in this sample
}

int ExChangePasswordCallBackFn(int av_index, const char *account, const char *old_password, const char *new_password, const char *new_iotc_authkey)
{
	//not implement in this sample
	return 0;
}

void ExAbilityRequestFn(int av_index, avServSendAbility send_ability)
{
	//not implement in this sample
}

//########################################################
//#     Enable / Disable live stream to AV client
//########################################################
void regedit_client_to_video(int SID, int avIndex)
{
	AV_Client *p = &gClientInfo[SID];
	p->avIndex = avIndex;
	p->bEnableLiveVideo = 1;
}

void unregedit_client_from_video(int SID)
{
	AV_Client *p = &gClientInfo[SID];
	p->bEnableLiveVideo = 0;
}

void regedit_client_stream_mode(int SID, int streamMode)
{
	AV_Client *p = &gClientInfo[SID];
	p->bTwoWayStream = streamMode;
}



//########################################################
//#     Handle client command
//########################################################
void Handle_IOCTRL_Cmd(int SID, int avIndex, char *buf, int type)
{
	printf("Handle CMD: ");
	switch (type)
	{
	case IOTYPE_USER_IPCAM_START:
	{
		SMsgAVIoctrlAVStream *p = (SMsgAVIoctrlAVStream *)buf;
		printf("IOTYPE_USER_IPCAM_START, ch:%d, avIndex:%d\n\n", p->channel, avIndex);
		int lock_ret = pthread_rwlock_wrlock(&gClientInfo[SID].sLock);
		if (lock_ret)
			printf("%s() Acquire SID %d rwlock error, ret = %d\n", __FUNCTION__, SID, lock_ret);

		regedit_client_to_video(SID, avIndex);

		lock_ret = pthread_rwlock_unlock(&gClientInfo[SID].sLock);
		if (lock_ret)
			printf("%s() Release SID %d rwlock error, ret = %d\n", __FUNCTION__, SID, lock_ret);

		printf("regedit_client_to_video OK\n");
	}
	break;
	case IOTYPE_USER_IPCAM_STOP:
	{
		SMsgAVIoctrlAVStream *p = (SMsgAVIoctrlAVStream *)buf;
		printf("IOTYPE_USER_IPCAM_STOP, ch:%d, avIndex:%d\n\n", p->channel, avIndex);
		int lock_ret = pthread_rwlock_wrlock(&gClientInfo[SID].sLock);
		if (lock_ret)
			printf("%s() Acquire SID %d rwlock error, ret = %d\n", __FUNCTION__, SID, lock_ret);

		unregedit_client_from_video(SID);
		//release lock
		lock_ret = pthread_rwlock_unlock(&gClientInfo[SID].sLock);

		if (lock_ret)
			printf("%s() Release SID %d rwlock error, ret = %d\n", __FUNCTION__, SID, lock_ret);
		printf("unregedit_client_from_video OK\n");
	}
	break;

	case IOTYPE_USER_IPCAM_AUDIOSTART:
	case IOTYPE_USER_IPCAM_AUDIOSTOP:
		//not implement in this sample
	break;
	case IOTYPE_USER_IPCAM_RECORD_PLAYCONTROL:
	{
		printf("IOTYPE_USER_IPCAM_RECORD_PLAYCONTROL, avIndex:%d\n", avIndex);
		Handle_IOTYPE_USER_IPCAM_RECORD_PLAYCONTROL(SID, avIndex, buf);
	}
	break;
	case IOTYPE_USER_IPCAM_DEVICE_SUPPORT_CLOUD_REQ:
	{
		SMsgAVIoctrlDeviceSupportOTAResp resp;
		resp.isSupport = VSAAS_CLOUD_SUPPORT_FLAG;
		int ret = avSendIOCtrl(avIndex, IOTYPE_USER_IPCAM_DEVICE_SUPPORT_CLOUD_RESP, (char *) &resp, sizeof(SMsgAVIoctrlDeviceSupportOTAResp));
		if(ret != AV_ER_NoERROR){
			printf("avSendIOCtrl fail[%d]\n", ret);
		}
	}
	break;
	default:
		printf("avIndex %d: non-handle type[%X]\n", avIndex, type);
		break;
	}
}

//########################################################
//#Thread - Start AV server and recv IOCtrl cmd for every new av idx
//########################################################
void *thread_ForAVServerStart(void *arg)
{
	int SID = *(int *)arg;
	free(arg);
	int ret;
	unsigned int ioType;
	char ioCtrlBuf[AV_MAX_IOCTRL_DATA_SIZE];
	struct st_SInfoEx Sinfo;

	printf("SID[%d], thread_ForAVServerStart, OK\n", SID);

	AVServStartInConfig avStartInConfig;
	AVServStartOutConfig avStartOutConfig;

	memset(&avStartInConfig, 0, sizeof(AVServStartInConfig));
	avStartInConfig.cb = sizeof(AVServStartInConfig);
	avStartInConfig.iotc_session_id = SID;
	avStartInConfig.iotc_channel_id = 0;
	avStartInConfig.timeout_sec = 30;
	avStartInConfig.password_auth = ExPasswordAuthCallBackFn;
	avStartInConfig.token_auth = ExTokenAuthCallBackFn;
	avStartInConfig.server_type = SERVTYPE_STREAM_SERVER;
	avStartInConfig.resend = ENABLE_RESEND;
	avStartInConfig.token_delete = ExTokenDeleteCallBackFn;
	avStartInConfig.token_request = ExTokenRequestCallBackFn;
	avStartInConfig.identity_array_request = ExGetIdentityArrayCallBackFn;
	avStartInConfig.change_password_request = ExChangePasswordCallBackFn;
	avStartInConfig.ability_request = ExAbilityRequestFn;

	if (ENABLE_DTLS)
		avStartInConfig.security_mode = AV_SECURITY_DTLS; // Enable DTLS, otherwise use AV_SECURITY_SIMPLE
	else
		avStartInConfig.security_mode = AV_SECURITY_SIMPLE;

	avStartOutConfig.cb = sizeof(AVServStartOutConfig);

	int avIndex = avServStartEx(&avStartInConfig, &avStartOutConfig);

	if (avIndex < 0)
	{
		printf("avServStartEx failed!! SID[%d] code[%d]!!!\n", SID, avIndex);
		printf("thread_ForAVServerStart: exit!! SID[%d]\n", SID);
		IOTC_Session_Close(SID);
		gOnlineNum--;
		pthread_exit(0);
	}

    memset(&Sinfo, 0, sizeof(struct st_SInfoEx));
    Sinfo.size = sizeof(struct st_SInfoEx);
    if (IOTC_Session_Check_Ex(SID, &Sinfo) == IOTC_ER_NoERROR)
	{
		char *mode[3] = {"P2P", "RLY", "LAN"};
		// print session information(not a must)
		if (isdigit(Sinfo.RemoteIP[0]))
			printf("Client is from[IP:%s, Port:%d] Mode[%s] VPG[%d:%d:%d] VER[%X] NAT[%d] AES[%d]\n", Sinfo.RemoteIP, Sinfo.RemotePort, mode[(int)Sinfo.Mode], Sinfo.VID, Sinfo.PID, Sinfo.GID, Sinfo.IOTCVersion, Sinfo.LocalNatType, Sinfo.isSecure);
	}
	printf("avServStartEx OK, avIndex[%d], resend[%d] two_way_streaming[%d]\n", avIndex, avStartOutConfig.resend, avStartOutConfig.two_way_streaming);
	printf("auth_type[%d] account_or_identity[%s]\n\n", avStartOutConfig.auth_type, avStartOutConfig.account_or_identity);
	regedit_client_stream_mode(SID, avStartOutConfig.two_way_streaming);
	avServSetResendSize(avIndex, 1024);

	while (1)
	{
		ret = avRecvIOCtrl(avIndex, &ioType, (char *)ioCtrlBuf, AV_MAX_IOCTRL_DATA_SIZE, 1000);
		if (ret >= 0)
		{
			Handle_IOCTRL_Cmd(SID, avIndex, ioCtrlBuf, ioType);
		}
		else if (ret != AV_ER_TIMEOUT)
		{
			printf("avIndex[%d], avRecvIOCtrl error, code[%d]\n", avIndex, ret);
			break;
		}
	}

	unregedit_client_from_video(SID);
	avServStop(avIndex);
	printf("SID[%d], avIndex[%d], thread_ForAVServerStart exit!!\n", SID, avIndex);

	IOTC_Session_Close(SID);
	gOnlineNum--;

	pthread_exit(0);
}

unsigned int getTimeStamp()
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return (tv.tv_sec * 1000 + tv.tv_usec / 1000);
}

//########################################################
//#Thread - IOTC Device Login
//########################################################
void *thread_Login(void *arg)
{
	int ret, sleep_count;
	DeviceLoginInput auth_option;

	printf("thread_Login start\n");

	while (gProcessRun)
	{
		auth_option.cb = sizeof(DeviceLoginInput);
		auth_option.authentication_type = AUTHENTICATE_BY_KEY;
		memcpy(auth_option.auth_key, gIotcAuthKey, IOTC_AUTH_KEY_LENGTH);

		ret = IOTC_Device_LoginEx((char *)arg, &auth_option);
		printf("IOTC_Device_LoginEx() ret = %d\n", ret);

		if (ret == IOTC_ER_NoERROR)
		{
			gIOTCLoginSuccess = 1;
			break;
		}
		else
		{
			PrintErrHandling(ret);
			for (sleep_count = 0; sleep_count < 5 && gProcessRun; sleep_count++)
				sleep(1);
		}
	}

	printf("thread_Login exit\n");
	pthread_exit(0);
}

//########################################################
//#    Thread - Send video frames(live) to regedited client
//########################################################
void *thread_VideoLiveFrame(void *arg)
{
	unsigned int totalCnt = 0;
	float hF = 0.0, lF = 0.0, totalFps = 0;
	long takeSec = 0, takeUSec = 0, sendFrameRoundTick = 0;
	int fpsCnt = 0, round = 0, framerate = FPS, sleepTick = 1000000 / framerate;
	int i = 0, avIndex = 0, bEnableLiveVideo = 0, bSendFrameOut = 0, size = 0, ret = 0, lock_ret = 0;
	char buf[VIDEO_BUF_SIZE];
	struct timeval tv, tv2;
	struct timeval tStart, tEnd;
	FILE *fp = NULL;
	FRAMEINFO_t frameInfo;
	fp = fopen(gLiveIframePath, "rb");
	if (fp == NULL)
	{
		printf("%s() Video File \'%s\' open error!!\n", __FUNCTION__, gLiveIframePath);
		printf("[Vidio] is DISABLED!!\n");
		printf("%s() exit\n", __FUNCTION__);
		pthread_exit(0);
	}

	// input file only one I frame for test
	size = fread(buf, 1, VIDEO_BUF_SIZE, fp);
	fclose(fp);
	if (size <= 0)
	{
		printf("%s() Video File \'%s\' read error!!\n", __FUNCTION__, gLiveIframePath);
		printf("[Vidio] is DISABLED!!\n");
		printf("%s() exit\n", __FUNCTION__);
		pthread_exit(0);
	}

	// *** set Video Frame info here ***
	memset(&frameInfo, 0, sizeof(FRAMEINFO_t));
	frameInfo.codec_id = MEDIA_CODEC_VIDEO_H264;
	frameInfo.flags = IPC_FRAME_FLAG_IFRAME;

	printf("%s() start OK\n", __FUNCTION__);

	while (gProcessRun)
	{
		frameInfo.timestamp = getTimeStamp();
		bSendFrameOut = 0;
		takeSec = 0, takeUSec = 0, sendFrameRoundTick = 0;

        FakeEventForVsaas((unsigned long)time(NULL));

		if (fpsCnt == 0)
			gettimeofday(&tv, NULL);

		for (i = 0; i < MAX_CLIENT_NUMBER; i++)
		{
			//get reader lock
			lock_ret = pthread_rwlock_rdlock(&gClientInfo[i].sLock);
			if (lock_ret)
				printf("Acquire SID %d rdlock error, ret = %d\n", i, lock_ret);

			avIndex = gClientInfo[i].avIndex;
			bEnableLiveVideo = gClientInfo[i].bEnableLiveVideo;

			//release reader lock
			lock_ret = pthread_rwlock_unlock(&gClientInfo[i].sLock);
			if (lock_ret)
				printf("Acquire SID %d rdlock error, ret = %d\n", i, lock_ret);

			if (avIndex < 0 || bEnableLiveVideo == 0)
			{
				continue;
			}

			// Send Video Frame to av-idx and know how many time it takes
			frameInfo.onlineNum = gOnlineNum;
			gettimeofday(&tStart, NULL);
			ret = avSendFrameData(avIndex, buf, size, &frameInfo, sizeof(FRAMEINFO_t));
			gettimeofday(&tEnd, NULL);

			takeSec = tEnd.tv_sec - tStart.tv_sec, takeUSec = tEnd.tv_usec - tStart.tv_usec;
			if (takeUSec < 0)
			{
				takeSec--;
				takeUSec += 1000000;
			}
			sendFrameRoundTick += takeUSec;
			totalCnt++;

			if (ret == AV_ER_EXCEED_MAX_SIZE) // means data not write to queue, send too slow, I want to skip it
			{
				usleep(5000);
				continue;
			}
			else if (ret == AV_ER_SESSION_CLOSE_BY_REMOTE)
			{
				printf("%s() AV_ER_SESSION_CLOSE_BY_REMOTE SID[%d]\n", __FUNCTION__, i);
				unregedit_client_from_video(i);
				continue;
			}
			else if (ret == AV_ER_REMOTE_TIMEOUT_DISCONNECT)
			{
				printf("%s() AV_ER_REMOTE_TIMEOUT_DISCONNECT SID[%d]\n", __FUNCTION__, i);
				unregedit_client_from_video(i);
				continue;
			}
			else if (ret == IOTC_ER_INVALID_SID)
			{
				printf("Session cant be used anymore\n");
				unregedit_client_from_video(i);
				continue;
			}
			else if (ret < 0)
				printf("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX\n");

			bSendFrameOut = 1;
		}

		if (bSendFrameOut && fpsCnt++ >= framerate)
		{
			round++;
			gettimeofday(&tv2, NULL);
			long sec = tv2.tv_sec - tv.tv_sec, usec = tv2.tv_usec - tv.tv_usec;
			if (usec < 0)
			{
				sec--;
				usec += 1000000;
			}
			usec += (sec * 1000000);

			long oneFrameUseTime = usec / fpsCnt;
			float fps = (float)1000000 / oneFrameUseTime;
			if (fps > hF)
				hF = fps;
			if (lF == 0.0)
				lF = fps;
			else if (fps < lF)
				lF = fps;
			printf("Fps = %f R[%d]\n", fps, round);
			fpsCnt = 0;
			totalFps += fps;
		}

		// notice the frames sending time for more specific frame rate control
		if (sleepTick > sendFrameRoundTick)
			usleep(sleepTick - sendFrameRoundTick);
	}

	printf("%s() exit High/Low [%f/%f] AVG[%f] totalCnt[%d]\n", __FUNCTION__, hF, lF, (float)totalFps / round, totalCnt);

	pthread_exit(0);
}

void create_livestream_thread()
{
	int ret;

	pthread_t ThreadVideoFrameData_ID;

	if ((ret = pthread_create(&ThreadVideoFrameData_ID, NULL, &thread_VideoLiveFrame, NULL)))
	{
		printf("pthread_create ret=%d\n", ret);
		exit(-1);
	}
	pthread_detach(ThreadVideoFrameData_ID);
}

//########################################################
//#  Initialize / Deinitialize client list of AV server
//########################################################
void InitAVInfo()
{
	int i;
	for (i = 0; i < MAX_CLIENT_NUMBER; i++)
	{
		memset(&gClientInfo[i], 0, sizeof(AV_Client));
		gClientInfo[i].avIndex = -1;
		gClientInfo[i].playBackCh = -1;
		pthread_rwlock_init(&(gClientInfo[i].sLock), NULL);
	}
	gIOTCLoginSuccess = -1;
}

void DeInitAVInfo()
{
	int i;
	for (i = 0; i < MAX_CLIENT_NUMBER; i++)
	{
        pthread_rwlock_destroy(&gClientInfo[i].sLock);
		memset(&gClientInfo[i], 0, sizeof(AV_Client));
		gClientInfo[i].avIndex = -1;
	}
}

//########################################################
//#     Callback function of IOTC login status
//########################################################
void LoginInfoCB(unsigned int nLoginInfo)
{
	if ((nLoginInfo & 0x04))
		printf("I can be connected via Internet\n");
}

// return
// -1 : invalid vsaasinfo
//  0 : new version vsaasinfo
//  1 : old version vsaasinfo, need convert to new format
int checkVsaasInfo(char *file_path, char *url, char *stoken){

    if(NULL == file_path || NULL == url || NULL == stoken)
        return -1;

    char vsaas_info[1024] = {0};
    int ret = 0;
    FILE *fp = fopen(file_path, "r");
    if(NULL == fp)
        return -1;

    ret = fread(vsaas_info, 1, sizeof(vsaas_info), fp);
    fclose(fp);

    if(ret > 0) {
        NebulaJsonObject *json_obj = NULL;
        const char *str_func = NULL;
        const char *str_url = NULL;
        const char *str_stoken = NULL;
        ret = Nebula_Json_Obj_Create_From_String(vsaas_info, &json_obj);
        if(!ret) {
            ret = Nebula_Json_Obj_Get_Sub_Obj_String(json_obj, "func", &str_func);
            if(!ret) {
                if(0 == strcmp(str_func, "setVsaasInfo")) {
                    const NebulaJsonObject *args_obj = NULL;
                    ret = Nebula_Json_Obj_Get_Sub_Obj(json_obj, "args", &args_obj);
                    if(!ret) {
                        ret = Nebula_Json_Obj_Get_Sub_Obj_String(args_obj, "url", &str_url);
                        if(!ret) {
                            strcpy(url, str_url);
                            ret = Nebula_Json_Obj_Get_Sub_Obj_String(args_obj, "stoken", &str_stoken);
                            if(!ret) {
                                strcpy(stoken, str_stoken);
                                ret = 1;
                            }
                        }
                    }
                } else if(0 == strcmp(str_func, "setCloudRecordingEndpoint")) {
                    ret = 0;
                } else {
                    ret = -1;
                }
            }
            Nebula_Json_Obj_Release(json_obj);
        }else{
            printf("[%s] Nebula_Json_Obj_Create_From_String fail, return : %d\n", __func__, ret);
        }
    }

    return (ret>=0)?ret:-1;
}

//########################################################
//#  Main function of this sample server
//########################################################
int main(int argc, char *argv[])
{
	int ret, SID;
	pthread_t ThreadLogin_ID;
	char vsaas_info_json[1024] = {0};
	FILE *vsaas_info_file_ptr = NULL;

	if (argc!=2)
	{
		printf("Argument Error!!!\n");
		printf("Usage: %s UID\n", argv[0]);
		return -1;
	}

    const char *license_key = getenv(ENV_LICENSE_KEY);
    if (license_key == NULL) {
        printf("Environment variable %s must be set!!\n", ENV_LICENSE_KEY);
        return -1;
    }
    ret = TUTK_SDK_Set_License_Key(license_key);
    if (ret < 0) {
        printf("TUTK_SDK_Set_License_Key: %d\n", ret);
        return -1;
    }

	strcpy(gUID, argv[1]);

	//set the output path of the log file when using log library.
	//AV_Set_Log_Path("/home/usr/log_file/av_server_log.txt", 0);

	do
	{
		InitAVInfo();

		create_livestream_thread();

		IOTC_Set_Max_Session_Number(MAX_CLIENT_NUMBER);
		// use which Master base on location, port 0 means to get a random port
		ret = IOTC_Initialize2(0);
		if (ret != IOTC_ER_NoERROR)
		{
			printf("  [] IOTC_Initialize2(), ret=[%d]\n", ret);
			PrintErrHandling(ret);
			DeInitAVInfo();
			return -1;
		}

        printf("IOTCAPI version[%s] AVAPI version[%s]\n", IOTC_Get_Version_String(), avGetAVApiVersionString());

		IOTC_Get_Login_Info_ByCallBackFn(LoginInfoCB);


        /* Convert an old-style format to the newest style format to achieve forward compatibility. */
        char url[128] = {0};
        char sToken[256] = {0};
        ret = checkVsaasInfo(gVsaasInfoFilePath, url, sToken);
        printf("checkVsaasInfo[%d]\n", ret);
        if(1 == ret){
            printf("url[%s], sToken[%s]\n", url, sToken);
            printf("Update vsaasinfo file\n");
            FILE *fp = fopen(gVsaasInfoFilePath, "w");
            char new_vsaas_info_json[1024] = {0};
            sprintf(new_vsaas_info_json, "{\"func\":\"setCloudRecordingEndpoint\", \"args\":{\"url\":\"https://%s/vsaas/api/v1/stream/stream_url/%s?stoken=%s\"}}", url, gUID, sToken);
            fwrite(new_vsaas_info_json , sizeof(char), strlen(new_vsaas_info_json), fp);
            fclose(fp);
        }

		//read vsaas info file && enable vsaas
		vsaas_info_file_ptr = fopen(gVsaasInfoFilePath, "r");
		if (vsaas_info_file_ptr == NULL)
		{
            ret = avEnableVSaaS(gUID, NULL, VsaasConfigChangedCb, VSaaSUpdateContractInfoCb);
        }
		else
		{

			gEnable_Vsaas = 1;
			ret = fread(vsaas_info_json, 1, 1024, vsaas_info_file_ptr);
			fclose(vsaas_info_file_ptr);
            ret = avEnableVSaaS(gUID, vsaas_info_json, VsaasConfigChangedCb, VSaaSUpdateContractInfoCb);
        }

		if (ret != AV_ER_NoERROR)
		{
            printf("avEnableVSaaS() error[%d]\n", ret);
            return -1;
		}else{
			printf("avEnableVSaaS() Successfully[%d]\n", ret);
		}

		avInitialize(MAX_CLIENT_NUMBER);

		// create thread to login because without WAN still can work on LAN
		if ((ret = pthread_create(&ThreadLogin_ID, NULL, &thread_Login, (void *)gUID)))
		{
			printf("Login Thread create fail, ret=[%d]\n", ret);
			return -1;
		}
		pthread_detach(ThreadLogin_ID);

		gProcessRun = 1;
		while (gProcessRun)
		{
			// Accept connection only when IOTC_Listen() calling
			SID = IOTC_Listen(1000);
			if (SID < 0)
			{
				PrintErrHandling(SID);
				if (SID == IOTC_ER_EXCEED_MAX_SESSION)
				{
					sleep(5);
				}
				continue;
			}

			int *sid = (int *)malloc(sizeof(int));
			*sid = SID;
			pthread_t Thread_ID;
			ret = pthread_create(&Thread_ID, NULL, &thread_ForAVServerStart, (void *)sid);
			if (ret < 0)
			{
				printf("pthread_create failed ret[%d]\n", ret);
			}
			else
			{
				pthread_detach(Thread_ID);
				gOnlineNum++;
			}
		}

		DeInitAVInfo();
		avDeInitialize();
		IOTC_DeInitialize();

		break;
	} while (1);

	return 0;
}
